/* File:        flrdisplay.P
**
** Author(s): Guizhen Yang
**
** Contact:   xsb-contact@cs.sunysb.edu
**
** Copyright (C) The Research Foundation of SUNY, 1999-2001
**
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
** 
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
** 
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
** 
** $Id: flrdisplay.P,v 1.4 2001/04/02 21:51:38 gzyang Exp $
** 
*/


:- compiler_options([xpp_on]).

#include "flora_terms.flh"
#include "standard.h"


:- import file_write/2 from xsb_writ.

:- export
	flora_stderr_write/1,
	flora_stdfdbk_write/1,
	flora_write/2.


/********************************************************************/
flora_stderr_write(X) :- flora_write(STDERR,X).
flora_stdfdbk_write(X) :- flora_write(STDFDBK,X).


/********************************************************************/
flora_write(IOPort,X) :-
	var(X),
	!,
	file_write(IOPort,X).

flora_write(IOPort,X) :-
	atomic(X),
	!,
	file_write(IOPort,X).

flora_write(IOPort,[]) :-
	!,
	file_write(IOPort,[]).

flora_write(IOPort,[H|T]) :-
	!,
	file_write(IOPort,'['),
	flora_write_args(IOPort,[H|T]),
	file_write(IOPort,']').

flora_write(IOPort,X) :-
	X =.. [WRAP_HILOG,Funct|Args],
	!,
	flora_write(IOPort,Funct),
	file_write(IOPort,'('),
	flora_write_args(IOPort,Args),
	file_write(IOPort,')').

flora_write(IOPort,WRAP_FDSKOLEM(O1,O2)) :-
	!,
	flora_write(IOPort,O1),
	file_write(IOPort,'.'),
	flora_write(IOPort,O2).

flora_write(IOPort,WRAP_IFDSKOLEM(O1,O2)) :-
	!,
	flora_write(IOPort,O1),
	file_write(IOPort,'!'),
	flora_write(IOPort,O2).

flora_write(IOPort,X) :-
	X =.. [Funct|Args],
	!,
	file_write(IOPort,Funct),
	file_write(IOPort,'('),
	flora_write_args(IOPort,Args),
	file_write(IOPort,')').


flora_write_args(_IOPort,[]) :- !.

flora_write_args(IOPort,[X]) :-
        !,
        flora_write(IOPort,X).

flora_write_args(IOPort,[X1,X2|Xs]) :-
        flora_write(IOPort,X1),
        file_write(IOPort,','),
        flora_write_args(IOPort,[X2|Xs]).
